//
// ukázka velmi jednoduchého diskrétního simulátoru v C++
// vim: set fileencoding=utf-8
//
// animace pohybu po obrazovce
// TODO - animovat postavu, doresit prekreslovani
//

#include <iostream>
#include <queue>        // priority_queue
#include <cstdlib>      // exit()

#include <unistd.h>     // usleep()

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// simulátor
/////////////////////////////////////////////////////////////////////////////

void error(const char *msg) {
    cerr << "ERROR: " << msg << endl;
    exit(1);
}

typedef void (*event_ptr_t) (); // ukazatel na funkci popisující událost

// Aktivační záznam události (pro jednoduchost bez priorit)
struct act_record {
    event_ptr_t event_ptr;      // odkaz na událost
    double atime;               // aktivační čas
    // konstruktor naplní položky:
    act_record(event_ptr_t e, double at): event_ptr(e), atime(at) {}
};

// Porovnání aktivačních záznamů z hlediska jejich pořadí v kalendáři.
// Je nutné pro uspořádání kalendáře typu priority_queue<act_record>
bool operator <(const act_record & a, const act_record & b)
{
    return a.atime > b.atime;  // POZOR HACK 
    // porovnání je obráceně -- lze opravit, ale 
    // musí se použít jiné než implicitní řazení u priority_queue
    // (menší čas ==> vyšší priorita při řazení)
}

// Jednoduchý kalendář událostí 
// - na začátku je vždy aktivační záznam s nejmenším časem
priority_queue<act_record> calendar;

// Modelový čas
double Time;

// Plánování události na zadaný čas
void schedule(event_ptr_t event_ptr, double at)
{
    //cout << "\t\t+Plánování na čas: " << at << endl;
    if (at < Time)
        error("Plánování do minulosti");
    act_record a(event_ptr, at);
    calendar.push(a);           // zařazení do kalendáře
}

/////////////////////////////////////////////////////////////////////////////
// popis modelu
/////////////////////////////////////////////////////////////////////////////

const double TSTART = 0;
const double TEND = 10000;

void event1()
{
    // popis činnosti
    char z = ".oOo"[unsigned(Time/100)%4];
    cout << '\b';
    if(Time<TEND/2) cout << ' ';
    cout << z << flush;
    schedule(event1, Time + 100);	// další výskyt
}

/////////////////////////////////////////////////////////////////////////////
// popis simulace
/////////////////////////////////////////////////////////////////////////////

int main()
{
    Time = TSTART;
    schedule(event1, Time);	// první aktivace
    cout << endl;
    cout << "*** začátek simulace:  čas = " << Time << endl;
    cout << endl;
    while (!calendar.empty()) {
	act_record a = calendar.top();	// přečteme první záznam
	calendar.pop();		// odstraníme první záznam
        if (a.atime > TEND) {
            Time = TEND;
            break;              // končíme simulaci
        }
        usleep(1000*long(a.atime-Time)); // "synchronizace" s reálným časem
	Time = a.atime;		// posuneme čas
	a.event_ptr();		// provedeme událost
    }
    cout << endl;
    cout << endl << "*** konec simulace:  čas = " << Time << endl;
    cout << endl;
}
